#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "lik.h"
#include "tree.h"
#include "calclik.h"

/* prototypes */

TNode *startNode;
double (*CalcLik)();

void Navigate(int, TNode *);


#define Ra R[0]
#define Rc R[1]
#define Rg R[2]
#define Rt R[3]

static double logNumCats;


/**********************************/	
void SetupLikelihood(TTree *tree)
{
	/* pre-calculate the log of the number of gamma categories */
	logNumCats=log(numCategories);
	InitFunc(globalMLE, tree->nodes);
	if (gammaRates)
		CalcLik=gammaCalcLik;
	else
		CalcLik=codonCalcLik;
}


/**********************************/	
void FinishLikelihood(TTree *tree)
{
	DoneFunc(tree->nodes);
}

/*************************************/
double codonCalcLik()
{
	static int k, cat;
	static double sum;
	static double *P, *Q, *R, *lut;
	static short *T;

	sum=0.0;
	for (cat=0; cat<numCategories; cat++) {
		Navigate(cat, startNode);

		lut=transMatrix[cat]+(startNode->branchNo << 4);
		P=patternLik[0];
		Q=startNode->Likelihood;
		R=startNode->branch->Likelihood;
		T=patternWeights[cat];

		for(k=0;k<numPatterns; k++) {
			if (*T) {
				*P= freqA*Ra*(lut[0]*Q[0] + lut[1]*Q[1] + lut[2]*Q[2] + lut[3]*Q[3]) +
					freqC*Rc*(lut[4]*Q[0] + lut[5]*Q[1] + lut[6]*Q[2] + lut[7]*Q[3]) +
					freqG*Rg*(lut[8]*Q[0] + lut[9]*Q[1] + lut[10]*Q[2] + lut[11]*Q[3]) +
					freqT*Rt*(lut[12]*Q[0] + lut[13]*Q[1] + lut[14]*Q[2] + lut[15]*Q[3]);

				sum+=log(*P)*(*T);
			}
			P++;
			T++;
			Q+=4;
			R+=4;
		}
	}
	
	return sum;
}


/*************************************/
double gammaCalcLik()
{
	static int k, cat;
	static double sum;
	static double *P, *Q, *R, *lut;
	static short *T;

	for (cat=0; cat<numCategories; cat++) {
		Navigate(cat, startNode);

		lut=transMatrix[cat]+(startNode->branchNo << 4);
		P=patternLik[cat];
		Q=startNode->Likelihood;
		R=startNode->branch->Likelihood;
		T=patternWeights[0];

		for(k=0; k<numPatterns; k++) {
			if (*T) {
				*P= freqA*Ra*(lut[0]*Q[0] + lut[1]*Q[1] + lut[2]*Q[2] + lut[3]*Q[3]) +
					freqC*Rc*(lut[4]*Q[0] + lut[5]*Q[1] + lut[6]*Q[2] + lut[7]*Q[3]) +
					freqG*Rg*(lut[8]*Q[0] + lut[9]*Q[1] + lut[10]*Q[2] + lut[11]*Q[3]) +
					freqT*Rt*(lut[12]*Q[0] + lut[13]*Q[1] + lut[14]*Q[2] + lut[15]*Q[3]);
			}
			P++;
			T++;
			Q+=4;
			R+=4;
		}
	}
	
	sum=0.0;

	/* sum across gamma rate categories */
	for (cat=1; cat<numCategories; cat++) {
		P=patternLik[0];
		Q=patternLik[cat];
		for(k=0; k<numPatterns; k++) {
			*P+=*Q;
			P++;
			Q++;
		}
	}
	
	/* then multiply the lot together (i.e. sum the logs) */
	P=patternLik[0];
	T=patternWeights[0];
	for(k=0; k<numPatterns; k++) {
		sum+=(log(*P)-logNumCats)*(*T);
		P++;
		T++;
	}
	
	return sum;
}


/*************************************/
void Navigate(int cat, TNode *posn)
{
	int k;
	double l,*P, *Q,*Q2, *lut,*lut2;
	short *T;
	TNode *newposn,*newposn2;

	newposn=posn->next->branch;		
	if(newposn->tipNo==-1)
		Navigate(cat, newposn);
	newposn2=posn->next->next->branch;
	if(newposn2->tipNo==-1)
		Navigate(cat, newposn2);

	P=posn->Likelihood;
	Q=newposn->Likelihood;
	Q2=newposn2->Likelihood;
	lut=transMatrix[cat]+(newposn->branchNo << 4);
	lut2=transMatrix[cat]+(newposn2->branchNo << 4);
	if (codonRates)
		T=patternWeights[cat];
	else
		T=patternWeights[0];
	for(k=0;k<numPatterns; k++) {
		if (T[k]) {
			P[0]=(lut[0]*Q[0] + lut[1]*Q[1] + lut[2]*Q[2] + lut[3]*Q[3])*(lut2[0]*Q2[0] + lut2[1]*Q2[1] + lut2[2]*Q2[2] + 
lut2[3]*Q2[3]);
			P[1]=(lut[4]*Q[0] + lut[5]*Q[1] + lut[6]*Q[2] + lut[7]*Q[3])*(lut2[4]*Q2[0] + lut2[5]*Q2[1] + lut2[6]*Q2[2] + 
lut2[7]*Q2[3]);
			P[2]=(lut[8]*Q[0] + lut[9]*Q[1] + lut[10]*Q[2] + lut[11]*Q[3])*(lut2[8]*Q2[0] + lut2[9]*Q2[1] + lut2[10]*Q2[2] + 
lut2[11]*Q2[3]);
			P[3]=(lut[12]*Q[0] + lut[13]*Q[1] + lut[14]*Q[2] + lut[15]*Q[3])*(lut2[12]*Q2[0] + lut2[13]*Q2[1] + 
lut2[14]*Q2[2] + lut2[15]*Q2[3]);
		}  
		P+=4;
		Q+=4;
		Q2+=4;
	}
}


